-----------------------------------------------------------------------------------------------
-- Client Lua Script for HarvestPlates
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
local HarvestPlates = {}

local nAddonVersion = 4

local sDefaultColor = "gray"

local tDefaultSettings = {
	bDebug = false,
	nAddonVersion = nAddonVersion,
	nDrawDistance = 150,
	nTimerRefreshSec = 0.3
}
 
function HarvestPlates:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
		
    return o
end

function HarvestPlates:Init()
    Apollo.RegisterAddon(self)
end
 
--- load ---

function HarvestPlates:OnLoad()
	self.tSettings = tDefaultSettings
	
	self.tNameplates = {}
	self.tUnitsBacklog = {}
	self.uPlayerUnit = nil
	self.nTargetId = nil
	
	self.sCurrentColor = sDefaultColor
	self.sFarmerChoice = ""
	
	self.uXml = XmlDoc.CreateFromFile("HarvestPlates.xml")
	
	Apollo.RegisterSlashCommand("harvest_color", "OnColorChange", self)
	Apollo.RegisterSlashCommand("harvest_farm", "OnFarmChange", self)
	
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
	Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)
	
	Apollo.CreateTimer("HarvestPlates_Refresh", self.tSettings.nTimerRefreshSec, true)
	Apollo.RegisterTimerHandler("HarvestPlates_Refresh", "OnTimerRefresh", self)
end

function HarvestPlates:OnSave(eLevel)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
	
	local tSettings = {}
	
	if self.sCurrentColor ~= sDefaultColor then
		tSettings.sCurrentColor = self.sCurrentColor
	end
	
	if self.sFarmerChoice ~= "" then
		tSettings.sFarmerChoice = self.sFarmerChoice
	end
	
	return tSettings
end

function HarvestPlates:OnRestore(eLevel, tData)
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
	
	if tData.sCurrentColor then
		self:OnColorChange(nil, tData.sCurrentColor)
	end
	
	if tData.sFarmerChoice then
		self:OnFarmChange(nil, tData.sFarmerChoice)
	end
end

function HarvestPlates:OnColorChange(sCommand, sValue)
	if sValue == nil or sValue == "" then
		sValue = sDefaultColor
	end
	for nUnitId, tUnitInfo in pairs(self.tNameplates) do
		tUnitInfo.window:FindChild("Type"):SetTextColor(sValue)
		tUnitInfo.window:FindChild("Name"):SetTextColor(sValue)
	end
	
	self.sCurrentColor = sValue
end

function HarvestPlates:OnFarmChange(sCommand, sValue)
	self.sFarmerChoice = sValue
end

function HarvestPlates:AddNameplate(uUnit)
	local nUnitId = uUnit:GetId()
	
	if self.tNameplates[nUnitId] ~= nil then
		self:debug(nUnitId .. " - nameplate already added")
		return
	end
	
	local uNameplate = Apollo.LoadForm(self.uXml, "Nameplate", "InWorldHudStratum", self)
	
	uNameplate:SetUnit(uUnit)
	
	local sTradeType = uUnit:GetHarvestRequiredTradeskillName()
	local wChildType = uNameplate:FindChild("Type")
	local wChildName = uNameplate:FindChild("Name")
	
	if sTradeType ~= "Farmer" then
		wChildType:SetText(string.format("%s (%s)", sTradeType, uUnit:GetHarvestRequiredTradeskillTier()))
	else
		wChildType:SetText(sTradeType)
	end
	wChildName:SetText(uUnit:GetName())
	
	wChildType:SetTextColor(self.sCurrentColor)
	wChildName:SetTextColor(self.sCurrentColor)
	
	self.tNameplates[nUnitId] = {
		window = uNameplate,
		unit = uUnit,
		isVisible = false,
		isMyTarget = false,
		isFarmerNode = (sTradeType == "Farmer")
	}
	
	self:UpdateNameplate(nUnitId)
end

function HarvestPlates:RemoveNameplate(nUnitId)
	if self.tNameplates[nUnitId] == nil then
		self:debug(nUnitId .. " - nameplate already removed")
		return
	end
	
	self.tNameplates[nUnitId].window:Destroy()
	self.tNameplates[nUnitId] = nil
end

function HarvestPlates:UpdateNameplate(nUnitId)
	local tData = self.tNameplates[nUnitId]
	local bIsVisible = self:ShouldNameplateBeVisible(nUnitId)
	
	if tData.isVisible == bIsVisible then
		return
	end
	
	tData.window:FindChild("FlickerProtector"):Show(bIsVisible)
	tData.isVisible = bIsVisible
end

function HarvestPlates:UpdateTargetStatus(nUnitId, bIsTarget)
	local tData = self.tNameplates[nUnitId]
	
	if tData == nil then
		return
	end
	
	tData.isMyTarget = bIsTarget
	self:UpdateNameplate(nUnitId)
end

function HarvestPlates:ShouldNameplateBeVisible(nUnitId)
	local tData = self.tNameplates[nUnitId]
	local bFarmVisible = true
	if tData.isFarmerNode and self.sFarmerChoice ~= "" and tData.unit:GetName() ~= self.sFarmerChoice then
		bFarmVisible = false
	end
		
	return bFarmVisible and tData.window:IsOnScreen() and not tData.window:IsOccluded() and (self:GetDistanceTo(tData.unit) <= self.tSettings.nDrawDistance or tData.isMyTarget)
end

function HarvestPlates:GetDistanceTo(uUnit)
	local tPos1 = self.uPlayerUnit:GetPosition()
	
	local tPos2 = uUnit:GetPosition()

	return math.sqrt(math.pow(tPos1.x - tPos2.x, 2) + math.pow(tPos1.y - tPos2.y, 2) + math.pow(tPos1.z - tPos2.z, 2))
end

function HarvestPlates:OnChangeWorld()
	self.uPlayerUnit = nil
end

function HarvestPlates:OnUnitCreated(uUnit)
	if self.uPlayerUnit == nil then
		self.tUnitsBacklog[uUnit:GetId()] = uUnit
		return
	end
	
	if uUnit:GetType() ~= "Harvest" or uUnit:CanBeHarvestedBy(self.uPlayerUnit) ~= true or self.tNameplates[uUnit:GetId()] ~= nil then
		return
	end	
	self:AddNameplate(uUnit)
end

function HarvestPlates:OnTimerRefresh()
	if self.uPlayerUnit == nil then
		self.uPlayerUnit = GameLib.GetPlayerUnit()
		
		if self.uPlayerUnit ~= nil then
			self:AddUnitsFromBacklog()
		end
		
		return
	end
	
	for nUnitId, tData in pairs(self.tNameplates) do
		self:UpdateNameplate(nUnitId)
	end
end

function HarvestPlates:OnUnitDestroyed(uUnit)
	if self.uPlayerUnit == nil then
		self.tUnitsBacklog[uUnit:GetId()] = nil
		return
	end
	
	if self.tNameplates[uUnit:GetId()] == nil then
		return
	end
	self:RemoveNameplate(uUnit:GetId())
end

function HarvestPlates:AddUnitsFromBacklog()
	for nUnitId, uUnit in pairs(self.tUnitsBacklog) do
		if uUnit:IsValid() then
			self:OnUnitCreated(uUnit)
		end
	end
	
	self.tUnitsBacklog = {}
end

function HarvestPlates:OnWorldLocationOnScreen(uHandler, uControl, bOnScreen)
	self:UpdateNameplate(uHandler:GetUnit():GetId())
end

function HarvestPlates:OnTargetUnitChanged(uUnit)
	if uUnit == nil then
		if self.nTargetId ~= nil then
			self:UpdateTargetStatus(self.nTargetId, false)
			self.nTargetId = nil
		end
		return
	end
	
	local nUnitId = uUnit:GetId()
	
	if self.nTargetId == nUnitId then
		return
	end
	
	if self.nTargetId ~= nil then
		self:UpdateTargetStatus(self.nTargetId, false)
	end
	
	self:UpdateTargetStatus(nUnitId, true)
	self.nTargetId = nUnitId
end

--- debug ---

function HarvestPlates:debug(...)
	if not self.tSettings.bDebug then
		return
	end
	
	if arg.n == 1 then
		Print(arg[1])
		return
	end
	
	for i, value in pairs(arg) do
		if i ~= "n" then
			Print(value)
		end
	end
end

function HarvestPlates:dump(name, value)
	SendVarToRover(name, value)
end

--- create object ---

local HarvestPlatesInst = HarvestPlates:new()
HarvestPlatesInst:Init()

